from . import linecache
from .coverageprocessor import determine_non_measured_lines, get_lines_in_file

FAILED_TESTS_KEY = 'failed_tests'
RECORDED_LINES_KEY = 'recorded_lines'


class Driver:
    def __init__(self, line_cache, file_hash_cache, initial_data):
        self.file_contents_cache = {}

        self.previously_failed_tests = frozenset()
        self.failed_tests = set()

        self.previously_recorded_lines = {}
        self.recorded_lines = {}

        self.line_cache = line_cache
        self.file_hash_cache = file_hash_cache

        if initial_data:
            self.previously_failed_tests = frozenset(
                initial_data.get(FAILED_TESTS_KEY, []))

            self.previously_recorded_lines = initial_data.get(
                RECORDED_LINES_KEY, {})

    def report_test_coverage(self, item_id, coverage_data):
        indices, non_measured_lines = determine_non_measured_lines(
            coverage_data, self.line_cache)

        test_lines = {
            filename: get_lines_in_file(
                filename,
                lines,
                self.file_contents_cache)
            for filename, lines in non_measured_lines.items()
            if lines
        }

        for filename, ranges in test_lines.items():
            filename_index = self.line_cache.filename_index(filename)
            for start, end, content in ranges:
                indices.append(
                    self.line_cache.save_record(
                        filename_index, start, end, content))

        assert item_id not in self.recorded_lines
        self.recorded_lines[item_id] = indices

    def report_test_failure(self, item_id):
        self.failed_tests.add(item_id)

    def should_execute_item(self, item_id, known_identical_items):
        if item_id not in self.previously_recorded_lines:
            return True

        if item_id in self.previously_failed_tests:
            return True

        old_file_data = self.previously_recorded_lines[item_id]

        for key in old_file_data:
            if key in known_identical_items:
                continue

            filename_index, start, end, content = self.line_cache.lookup(key)
            filename = self.line_cache.filenames[filename_index]

            if self.file_hash_cache.is_identical(filename):
                known_identical_items.add(key)
                continue

            new_line_data = get_lines_in_file(
                filename,
                range(start, end),
                self.file_contents_cache)

            if len(new_line_data) != 1:
                return True

            _, _, expected_content = new_line_data[0]

            if content != linecache.hash(expected_content):
                return True

        return False

    def to_json(self):
        line_data = {}
        line_data.update(self.previously_recorded_lines)
        line_data.update(self.recorded_lines)

        return {
            FAILED_TESTS_KEY: list(self.failed_tests),
            RECORDED_LINES_KEY: line_data,
        }



import coverage

try:
    import ujson
except ImportError:
    import json as ujson

from . import linecache, filehashcache, driver

CACHE_KEY = 'cache/coverage-by-test'
CACHE_VERSION_KEY = 'version'
CACHE_VERSION = 5
CACHE_FAILED_TESTS = 'failed_tests'
CACHE_FILE_HASH_CACHE_KEY = 'file_hashes'
CACHE_LINE_CACHE_KEY = 'line_cache'
CACHE_DRIVER_KEY = 'driver'


class CoverageExclusionPlugin:
    def __init__(self, config):
        assert config.cache

        self.config = config
        self.current_cov = None

        cache_data = ujson.loads(config.cache.get(CACHE_KEY, '{}'))
        if cache_data.get(CACHE_VERSION_KEY) != CACHE_VERSION:
            cache_data = {}

        self.file_hash_cache = filehashcache.FileHashCache(
            cache_data.get(CACHE_FILE_HASH_CACHE_KEY))

        self.line_cache = linecache.LineCache(
            cache_data.get(CACHE_LINE_CACHE_KEY))

        self.driver = driver.Driver(
            self.line_cache,
            self.file_hash_cache,
            cache_data.get(CACHE_DRIVER_KEY))

    def pytest_runtest_setup(self, item):
        assert not self.current_cov

        self.current_cov = coverage.Coverage()
        self.current_cov.start()

    def pytest_runtest_teardown(self, item, nextitem):
        if not self.current_cov:
            return

        self.current_cov.stop()

        data = self.current_cov.get_data()
        self.current_cov = None

        data.update(item._extra_cov_data)

        self.driver.report_test_coverage(item.nodeid, data)

    def pytest_runtest_logreport(self, report):
        if report.failed and 'xfail' not in report.keywords:
            self.driver.report_test_failure(report.nodeid)

    def pytest_sessionfinish(self, session):
        self.file_hash_cache.hash_missing_files(self.line_cache.filenames)

        self.config.cache.set(CACHE_KEY, ujson.dumps({
            CACHE_VERSION_KEY: CACHE_VERSION,
            CACHE_FILE_HASH_CACHE_KEY: self.file_hash_cache.to_json(),
            CACHE_LINE_CACHE_KEY: self.line_cache.to_json(),
            CACHE_DRIVER_KEY: self.driver.to_json(),
        }))

    def pytest_collection_modifyitems(self, session, config, items):
        to_keep = []
        to_skip = []

        known_identical_items = set()

        for item in items:
            if self._should_execute_item(item, known_identical_items):
                to_keep.append(item)
            else:
                to_skip.append(item)

        items[:] = to_keep
        config.hook.pytest_deselected(items=to_skip)

    def pytest_collectstart(self, collector):
        self.collect_cov = coverage.Coverage()
        self.collect_cov.start()

    def pytest_itemcollected(self, item):
        self.collect_cov.stop()
        item._extra_cov_data = self.collect_cov.get_data()

    def _should_execute_item(self, item, known_identical_items):
        if item.get_marker('external_dependencies'):
            return True

        return self.driver.should_execute_item(
            item.nodeid, known_identical_items)


def pytest_configure(config):
    config.pluginmanager.register(
        CoverageExclusionPlugin(config),
        "coverage-exclusion")


def _debug_coverage_data(data):
    for filename in data.measured_files():
        if 'site-packages' in filename:
            continue

        print('{}: {}'.format(filename, data.lines(filename)))


def _debug_lines_in_file(filename, data):
    if 'site-packages' in filename:
        return

    print('{}: {}'.format(filename, data))



import os.path
import platform

from setuptools import setup, find_packages

HERE = os.path.dirname(__file__)


def read_version():
    about_data = {}

    with open(os.path.join(HERE, 'covexclude/__about__.py')) as fp:
        exec(fp.read(), about_data)

    return about_data['__version__']


with open(os.path.join(HERE, 'README.rst')) as fp:
    long_description = fp.read()


install_requires = [
    'coverage>=4.0.0,<5.0.0',
]

if platform.python_implementation() != 'PyPy':
    install_requires += [
        'ujson>=1.0,<2.0',
    ]

setup(
    name='pytest-cov-exclude',
    version=read_version(),
    description='Pytest plugin for excluding tests based on coverage data',
    long_description=long_description,

    author='Magnus Hallin',
    author_email='mhallin@fastmail.com',

    url='https://github.com/mhallin/cov-exclude-py',

    license='MIT',

    packages=find_packages(exclude=['tests']),

    package_data={
        '': ['LICENSE', '*.rst'],
    },

    entry_points={
        'pytest11': [
            'cov-exclude = covexclude.plugin',
        ],
    },

    install_requires=install_requires,

    extras_require={
        'dev': [
            'pytest>=2.8.0,<2.9.0',
        ],
        'dist': [
            'twine',
            'wheel',
        ],
    },

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Testing',
    ],

    keywords=[
        'cover',
        'coverage',
        'pytest',
        'py.test',
        'performance',
        'speed',
    ]
)
